"""
WebSocket服务器
1. 提交json格式，返回也json格式。可在浏览器中用JS调用，或通过其它语言访问
2. 提交pickle格式，返回也json格式。内部调用

使用FastAPI搭建的WebSocket服务
1. pip install uvicorn[standard], 必须指定stanadard，否则安装后不支持websocket服务
2. `/ws/json` 可用其它语言调用
3. `/ws/bytes` 内部调用
4. `/ws/client` `/ws/amdin`, 反弹RPC

"""
from datetime import datetime
from typing import List, Any, Dict, Union

from IPy import IP
from fastapi import WebSocket, WebSocketDisconnect, Depends, Query
from fastapi import status
from fastapi.security.utils import get_authorization_scheme_param
from loguru import logger

from .app_ import app
from ..caller import call, before_call
from ..model import Format, RspModel
from ..serializer.json_ import obj_to_dict, dict_to_json
from ..serializer.pkl_gzip import deserialize, serialize
from ..utils.check_ import check_ip


async def get_current_user(ws: WebSocket, token: Union[str, None] = Query(None)):
    from ..config import AUTH_CHECK, AUTH_TOKENS
    if not AUTH_CHECK:
        return 'anonymous'
    authorization: str = ws.headers.get("Authorization")
    scheme, param = get_authorization_scheme_param(authorization)
    # 如果从头取不到token就从query中取
    if param:
        token = param
    return AUTH_TOKENS.get(token, None)


class ConnectionManager:
    def __init__(self):
        self.active_connections: List[WebSocket] = []

    async def connect(self, ws: WebSocket):
        await ws.accept()
        self.active_connections.append(ws)

    def disconnect(self, ws: WebSocket):
        self.active_connections.remove(ws)

    @staticmethod
    async def send_personal_message(message: str, ws: WebSocket):
        await ws.send_text(message)

    async def broadcast(self, message: str):
        for connection in self.active_connections:
            await connection.send_text(message)


manager = ConnectionManager()


async def _do(ws: WebSocket,
              func: str,

              args: List[Any] = [],
              kwargs: Dict[str, Any] = {},
              fmt: Format = Format.JSON,
              cache_get: bool = True, cache_expire: int = 86400,
              async_remote: bool = True,
              user: str = None,  # 没有用到，用于token认证
              ):
    """实际处理函数"""
    try:
        before_call(ws.client.host, user, func)
        key, buf, data = await call(user, func, args, kwargs, cache_get, cache_expire, async_remote)
    except Exception as e:
        # 主要是处理
        key = type(e).__name__
        # 这里没有缓存，因为这个错误是服务器内部检查
        data = RspModel(status=status.HTTP_401_UNAUTHORIZED,
                        datetime=datetime.now().isoformat(),
                        func=func, args=args, kwargs=kwargs)
        data.type = type(e).__name__
        data.data = repr(e)
        data = data.dict()
        buf = serialize(data).read()

    if fmt == Format.PKL_GZIP:
        return buf

    if data is None:
        data = deserialize(buf)

    # DataFrame需要csv格式时走此路径
    if fmt == Format.CSV and data['type'] in ('DataFrame', 'Series'):
        return data['data'].to_csv()

    # JSON格式时，为了方便DataFrame的显示，做一下转换
    data['data'] = obj_to_dict(data['data'])

    return data


@app.websocket("/ws/json")
async def websocket_endpoint_json(websocket: WebSocket, user=Depends(get_current_user)):
    """处理JSON，发送JSON或csv"""
    await manager.connect(websocket)
    try:
        while True:
            req = await websocket.receive_json(mode="text")
            logger.info(req)
            rsp = await _do(websocket, **req, user=user)
            # 部分类型换转json有问题，所以使用特殊的转换函数
            await websocket.send_text(dict_to_json(rsp))
    except WebSocketDisconnect:
        manager.disconnect(websocket)


@app.websocket("/ws/bytes")
async def websocket_endpoint_bytes(websocket: WebSocket, user=Depends(get_current_user)):
    """处理pkl.gz发送pkl.gz"""
    await manager.connect(websocket)
    try:
        while True:
            req = await websocket.receive_bytes()
            req = deserialize(req)
            logger.info(req)
            req['fmt'] = Format.PKL_GZIP
            rsp = await _do(websocket, **req, user=user)
            await websocket.send_bytes(rsp)
    except WebSocketDisconnect:
        manager.disconnect(websocket)


# 房间，分存客户端和管理端，中转时用于查找对应的ws
rooms_client: Dict[str, WebSocket] = {}
rooms_admin: Dict[str, WebSocket] = {}


@app.websocket("/ws/client")
async def websocket_endpoint_client(websocket: WebSocket, room: str, user=Depends(get_current_user)):
    """被控端。肉鸡接入后，等待控制端接入，然后转发到肉鸡，然后转发。必须要进入同一房间"""
    await manager.connect(websocket)
    try:
        if user is None:
            raise Exception(f"Unauthorized")

        rooms_client[room] = websocket
        while True:
            req = await websocket.receive_bytes()
            # 可以将被控端的二进制解码，然后转成json，实现跨语言，这里先不处理
            await rooms_admin[room].send_bytes(req)

    except WebSocketDisconnect:
        manager.disconnect(websocket)
    except Exception as e:
        logger.error(e)
        manager.disconnect(websocket)
        await websocket.close()


@app.websocket("/ws/admin")
async def websocket_endpoint_admin(websocket: WebSocket, room: str, user=Depends(get_current_user)):
    """控制端。等待控制端接入，然后转发"""
    await manager.connect(websocket)
    try:
        # 确保连上admin的用户有权限
        from ..config import IP_CHECK, IP_ALLOW, IP_BLOCK

        if IP_CHECK:
            host = IP(websocket.client.host)
            if not check_ip(IP_ALLOW, host, False):
                raise Exception(f'IP Not Allowed, {host} not in allowlist')
            if not check_ip(IP_BLOCK, host, True):
                raise Exception(f'IP Not Allowed, {host} in blocklist')

        if user is None:
            raise Exception(f"Unauthorized")

        rooms_admin[room] = websocket
        while True:
            req = await websocket.receive_bytes()
            await rooms_client[room].send_bytes(req)

    except WebSocketDisconnect:
        manager.disconnect(websocket)
    except Exception as e:
        logger.error(e)
        manager.disconnect(websocket)
        await websocket.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-05-15
"""
缓存
将查询的结果缓存起来，减少对外部数据源的压力，减少限额消耗
"""
from .config import CACHE_TYPE

# 测试时使用的内存redis
if CACHE_TYPE == 'fakeredis':
    import fakeredis

    cache = fakeredis.FakeStrictRedis()
    cache_is_async = False

# 异步版Redis
if CACHE_TYPE == 'aioredis':
    import aioredis
    from .config import REDIS_URL

    cache = aioredis.StrictRedis.from_url(REDIS_URL)
    cache_is_async = True


async def async_cache_get(name):
    """统一缓存读取方式"""
    if cache_is_async:
        return await cache.get(name)
    # 阻塞方式调用
    return cache.get(name)


async def async_cache_setex(name, time, value):
    """统一缓存写入方式"""
    if cache_is_async:
        return await cache.setex(name, time, value)
    # 阻塞方式调用
    cache.setex(name, time, value)


async def async_cache_incrby(name, amount):
    """统一缓存写入方式"""
    if cache_is_async:
        return await cache.incrby(name, amount)
    # 阻塞方式调用
    cache.incrby(name, amount)


async def async_cache_decrby(name, amount):
    """统一缓存写入方式"""
    if cache_is_async:
        return await cache.decrby(name, amount)
    # 阻塞方式调用
    cache.decrby(name, amount)


async def async_cache_keys(pattern):
    """统一缓存读取方式"""
    if cache_is_async:
        return await cache.keys(pattern)
    # 阻塞方式调用
    return cache.keys(pattern)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-05-16

"""
使用wesockets实现的客户端
1. 默认提交pickle格式，返回也是pickle格式
2. 原接口是异步版，必须使用async with，封装成了同步版with
3. 为了使用方便，还提供了不需with的版本，可直接调用

"""
from datetime import datetime
from urllib.parse import urlparse

from loguru import logger

from ..caller import call
from ..model import Format, RspModel
from ..serializer.json_ import dict_to_json, json_to_dict, dict_to_obj
from ..serializer.pkl_gzip import serialize, deserialize
from ..utils.async_ import to_sync
from ..utils.check_ import check_methods


def process_response(r):
    """处理响应。根据不同的响应分别处理

    Parameters
    ----------
    r: Response

    Returns
    -------
    object
    json
    csv

    """
    if isinstance(r, bytes):
        # 二进制，反序列化后是字典
        data = deserialize(r)
        if data['status'] == 200:
            return data['data']
        return data
    else:
        # json字符串转成字字典
        data = json_to_dict(r)
        # 特殊类型先还原
        if data.get('type', None) in ('DataFrame', 'Series', 'ndarray'):
            data['data'] = dict_to_obj(data['data'])
        # 成功响应直接返回数据区
        if data['status'] == 200:
            return data['data']
        return data


class WebSocketConnection:
    def __init__(self, url: str, token=None):
        path = urlparse(url).path
        assert path.endswith(('/json', '/bytes', '/client', '/admin')), 'Python语言优先使用bytes，其它语言使用json'

        if path.endswith('/json'):
            self._fmt = Format.JSON
        else:
            self._fmt = Format.PKL_GZIP

        self._url = url
        self._token = token
        self._with = False  # 记录是否用了with

    async def __aenter__(self):
        """异步async with"""
        from websockets import connect

        headers = None if self._token is None else {"Authorization": f"Bearer {self._token}"}
        self._client = connect(self._url, extra_headers=headers)

        self._ws = await self._client.__aenter__()
        self._with = True
        return self

    async def __aexit__(self, exc_type=None, exc_value=None, traceback=None):
        """异步async with"""
        await self._client.__aexit__(exc_type, exc_value, traceback)

    def __enter__(self):
        """同步with"""
        to_sync(self.__aenter__)()
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        """同步with"""
        to_sync(self.__aexit__)()

    async def call(self, func, args, kwargs,
                   fmt: Format = Format.PKL_GZIP,
                   cache_get: bool = True, cache_expire: int = 3600,
                   async_remote=True):
        # 还没有执行过with就内部主动执行一次
        if not self._with:
            await self.__aenter__()

        # 如果是JSON格式可以考虑返回CSV
        rsp_fmt = self._fmt

        d = dict(func=func, args=args, kwargs=kwargs,
                 fmt=rsp_fmt.value,
                 cache_get=cache_get, cache_expire=cache_expire, async_remote=async_remote)

        if self._fmt == Format.PKL_GZIP:
            # 二进制格式
            await self._ws.send(serialize(d).read())
        else:
            # json格式
            await self._ws.send(dict_to_json(d))

        rsp = await self._ws.recv()
        return process_response(rsp)

    async def reverse(self):
        """反弹RPC的被控端"""
        while True:
            req = await self._ws.recv()
            req = deserialize(req)
            logger.info(req)
            # 不需要缓存
            req['cache_get'] = False
            req['cache_expire'] = 0
            func = req['func']
            try:
                from ..config import METHODS_ALLOW, METHODS_BLOCK, METHODS_CHECK

                # 反弹RPC权限要进行限制
                if METHODS_CHECK:
                    methods = func.split('.')
                    if not check_methods(METHODS_ALLOW, methods, False):
                        raise Exception(f'Method Not Allowed, {func} not in allowlist')
                    if not check_methods(METHODS_BLOCK, methods, True):
                        raise Exception(f'Method Not Allowed, {func} in blocklist')

                key, buf, data = await call(**req)
            except Exception as e:
                # 主要是处理
                key = type(e).__name__
                # 这里没有缓存，因为这个错误是服务器内部检查
                data = RspModel(status=401,  # status.HTTP_401_UNAUTHORIZED,
                                datetime=datetime.now().isoformat(),
                                func=func, args=req['args'], kwargs=req['args'])
                data.type = type(e).__name__
                data.data = repr(e)
                data = data.dict()
                buf = serialize(data).read()

            await self._ws.send(buf)



"""
演示函数
"""
import asyncio
import time

import pandas as pd


def sync_say_hi(name):
    time.sleep(1)
    return f'Hello {name}'


async def async_say_hi(name):
    await asyncio.sleep(1)
    return f'Hello {name}'


def div(a, b):
    return a / b


def test():
    pd._testing._N = 250
    return pd._testing.makeTimeDataFrame()


## 以下为python 3.6下module不支持__getattr__的处理方法，请按实际情况进行设置
import sys


class Wrapper(object):
    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __getattr__(self, name):
        # Perform custom logic here
        try:
            return getattr(self.wrapped, name)
        except AttributeError as e:
            raise e


sys.modules[__name__] = Wrapper(sys.modules[__name__])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-05-15

"""
HTTP客户端
默认提交pickle格式，返回也是pickle格式

1. RequestsConnection，使用requests实现。原本只支持同步，但此处已经改成支持异步
2. HttpxConnection，使用httpx实现。使用异步，但也可以指定成同步使用

"""
from io import BytesIO
from urllib.parse import urlparse

import pandas as pd

from ..model import Format
from ..serializer.json_ import dict_to_obj
from ..serializer.pkl_gzip import deserialize, serialize
from ..utils.async_ import to_sync


def process_response(r):
    """处理HTTP响应。根据不同的响应分别处理

    Parameters
    ----------
    r: Response

    Returns
    -------
    object
    json
    csv

    """
    if r.status_code != 200:
        raise Exception(f'{r.status_code}, {r.text}')
    content_type = r.headers['content-type']
    if content_type.startswith('application/octet-stream'):
        data = deserialize(r.content)
        if data['status'] == 200:
            return data['data']
        return data
    elif content_type.startswith('application/json'):
        data = r.json()
        if data.get('type', None) in ('DataFrame', 'Series', 'ndarray'):
            return dict_to_obj(data['data'])
        if data['status'] == 200:
            return data['data']
        return data
    elif content_type.startswith('text/plain'):
        # 纯文本，表示返回的CSV格式
        return pd.read_csv(BytesIO(r.content), index_col=0)


class HttpxConnection:
    """使用httpx实现的客户端连接支持同步和异步"""
    # 超时，请求超求和响应超时，秒
    timeout = (5, 30)

    def __init__(self, url, token=None):
        import httpx

        path = urlparse(url).path
        assert path.endswith(('/get', '/post', '/file')), 'Python语言优先使用file，其它语言使用post'
        if path.endswith('/file'):
            self._fmt = Format.PKL_GZIP
        else:
            self._fmt = Format.JSON

        self._url = url
        self._token = token
        self._client = httpx.AsyncClient()

    async def __aenter__(self):
        """异步async with"""
        await self._client.__aenter__()
        return self

    async def __aexit__(self, exc_type=None, exc_value=None, traceback=None):
        """异步async with"""
        await self._client.__aexit__(exc_type, exc_value, traceback)

    def __enter__(self):
        """同步with"""
        to_sync(self._client.__aenter__)()
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        """同步with"""
        to_sync(self._client.__aexit__)()

    async def call(self, func, args, kwargs,
                   fmt: Format = Format.PKL_GZIP,
                   cache_get: bool = True, cache_expire: int = 3600,
                   async_remote=True):
        """调用函数

        Parameters
        ----------
        func: str
            函数全名
        args: tuple
            函数位置参数
        kwargs: dict
            函数命名参数
        fmt: Format
            指定响应格式
        cache_get: bool
            是否优先从缓存中获取
        cache_expire: int
            指定缓存超时。超时此时间将过期，指定0表示不进行缓存
        async_remote: bool
            异步方式调用

        """
        # 如果是JSON格式可以考虑返回CSV
        rsp_fmt = self._fmt

        # httpx解析枚举有问题，只能提前转成value，而requests没有此问题
        params = dict(func=func, fmt=rsp_fmt.value,
                      cache_get=cache_get, cache_expire=cache_expire, async_remote=async_remote)
        data = {'args': args, 'kwargs': kwargs}
        headers = None if self._token is None else {"Authorization": f"Bearer {self._token}"}

        if self._fmt == Format.PKL_GZIP:
            files = {"file": serialize(data).read()}
            r = await self._client.post(self._url, headers=headers, params=params, timeout=self.timeout, files=files)
        elif self._fmt == Format.JSON:
            r = await self._client.post(self._url, headers=headers, params=params, timeout=self.timeout, json=data)

        return process_response(r)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-05-16

"""
同步与异步互转的小工具

FastAPI大量使用了async异步编程，如果内有阻塞函数或执行过久，将无法处理新请求
而大部分API都是同步阻塞模式，所以需要转异步

而在客户端，为了达到与原API一样，所以还需要有异步转同步功能

某些情况下需要nest_asyncio补丁才不会报错
"""
import asyncio
import inspect
from functools import partial, wraps

# 可能出现RuntimeError: This event loop is already running，所以打补丁
import nest_asyncio

nest_asyncio.apply()


def to_async(f):
    """同步转异步"""

    @wraps(f)
    async def decorated(*args, **kwargs):
        if inspect.iscoroutinefunction(f):
            # 已经是异步函数了就直接调用
            return await f(*args, **kwargs)

        # python 3.6下没有get_running_loop，只能直接复制代码
        loop = asyncio._get_running_loop()
        if loop is None:
            raise RuntimeError('no running event loop')

        return await loop.run_in_executor(None, partial(f, *args, **kwargs))

    return decorated


def to_sync(f):
    """异步转同步"""

    @wraps(f)
    def decorated(*args, **kwargs):
        if not inspect.iscoroutinefunction(f):
            # 已经是同步函数了就直接调用
            return f(*args, **kwargs)

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            # RuntimeError: There is no current event loop in thread 'ThreadPoolExecutor-0_0'.
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # RuntimeError: This event loop is already running，某些情况下，需要nest_asyncio配合
        return loop.run_until_complete(f(*args, **kwargs))

    return decorated


if __name__ == '__main__':
    import time


    @to_async
    def do_sync_work(name: str):
        time.sleep(1)
        return f"同步 + 上加装饰器, {name}"


    async def do_async_work(name: str):
        await asyncio.sleep(1)
        return f"异步 + 不加装饰器, {name}"


    async def async_main():
        # 装饰器通常用法
        message = await do_sync_work(name="改异步")
        print(message)


    def sync_main():
        # 装饰器特殊用法
        message = to_sync(do_async_work)(name="改同步")
        print(message)


    asyncio.run(async_main())
    sync_main()


    async def async_main_2():
        # 套娃
        message = await to_async(to_sync(do_sync_work))(name="改")
        print(message)


    def sync_main_2():
        # 套娃
        message = to_sync(to_async(to_sync(do_sync_work)))(name="改")
        print(message)


    asyncio.run(async_main_2())
    sync_main_2()



import eikon as ek

from ..config import EIKON_APP_KEY

ek.set_app_key(EIKON_APP_KEY)
# 用完后清空，防止被客户端获取
del EIKON_APP_KEY


def set_app_key(app_key):
    pass


def get_app_key():
    return ""


def __getattr__(name):
    return ek.__getattr__(name)



"""
jqdatasdk客户端替换
1. 将底层的instance替换成RpcClient。实现数据库相关的操作也能操作
"""


def hack(client):
    # 内部类就可以了，不用暴露出去
    class HackClient:
        def __init__(self, c):
            self.client = c
            self.not_auth = False

        def __getattr__(self, item):
            return self.client.__getattr__(item)

        def __call__(self, *args, **kwargs):
            return self

        def _reset(self):
            pass

        def ensure_auth(self):
            pass

        def logout(self):
            pass

    # 最简洁的用法，直接将最低层取数据的部分替换
    import jqdatasdk.client
    jqdatasdk.client.JQDataClient.instance = HackClient(client)



"""
TuShare客户端替换
1. 将TuShare底层的pro替换成RpcClient
2. set_token和pro_api还是使用的客户本地API
"""


def hack(client):
    # 最简洁的用法，直接将最低层取数据的部分替换
    def __pro_api(token='', timeout=30):
        return client

    import tushare
    tushare.pro_api = __pro_api



# 应当用计划任务，每天凌晨0点执行
# 每天按需求重值部分用户的配额信息，第二天即可使用
import asyncio

from ksrpc.cache import async_cache_setex, async_cache_get, async_cache_keys


async def test():
    user = 'nobody'
    module = 'demo'
    func_name = 'demo.test'

    m_quota_key = f'QUOTA/{user}/{module}'
    f_quota_key = f'QUOTA/{user}/{func_name}'
    await async_cache_setex(m_quota_key, 86400 * 7, 100)
    await async_cache_setex(f_quota_key, 86400 * 7, 200)


async def async_main():
    keys = await async_cache_keys('QUOTA*')
    for key in keys:
        print(await async_cache_get(key))
    for key in keys:
        await async_cache_setex(key, 86400 * 7, 0)
    for key in keys:
        print(await async_cache_get(key))


if __name__ == '__main__':
    asyncio.run(test())
    asyncio.run(async_main())



"""
HTTP服务端
1. 提交json，返回csv或json。通过浏览器访问，或通过其它语言，可以走此通道
2. 提交pickle，返回pickle。python语言限定，这种方法更强大

通过FastAPI搭建的服务
1. 大量使用异步， 所以需要被调用的API也能转成异步，否则会阻塞之后的请求
2. 其它语言用户请走`/api/get`或`/api/post`
3. `/api/file`为python内部使用

"""
from datetime import datetime
from typing import Dict, Any, List

from fastapi import Query, Body, File, Depends
from fastapi import status
from fastapi.requests import Request
from fastapi.responses import PlainTextResponse, StreamingResponse
from fastapi.security import OAuth2PasswordBearer

from .app_ import app
from ..caller import call, before_call
from ..model import Format, RspModel
from ..serializer.json_ import obj_to_dict
from ..serializer.pkl_gzip import deserialize, serialize

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token", auto_error=False)


async def get_current_user(token: str = Depends(oauth2_scheme)):
    from ..config import AUTH_TOKENS, AUTH_CHECK

    if not AUTH_CHECK:
        return 'anonymous'
    return AUTH_TOKENS.get(token, None)


@app.get("/api/get")
async def api_get(request: Request,
                  func: str = Query(..., min_length=1),

                  fmt: Format = Query(Format.CSV),
                  cache_get: bool = Query(True), cache_expire: int = Query(86400, lt=86400 * 15),
                  async_remote: bool = Query(True),
                  user: str = Depends(get_current_user),
                  ):
    """无参数函数。可以通过get发起请求。可在浏览器中直接发起"""
    return await _do(**locals())


@app.post("/api/post")
async def api_post(request: Request,
                   args: List[Any] = Body([]),
                   kwargs: Dict[str, Any] = Body({}),

                   func: str = Query(..., min_length=1),

                   fmt: Format = Query(Format.CSV),
                   cache_get: bool = Query(True), cache_expire: int = Query(86400, lt=86400 * 15),
                   async_remote: bool = Query(True),
                   user: str = Depends(get_current_user),
                   ):
    """简单参数函数。可通过post请求，使用body区传json参数。可跨语言。可用Postman发起"""
    return await _do(**locals())


@app.post("/api/file")
async def api_file(request: Request,
                   file: bytes = File(...),  # file要写在最前面，否则报错

                   func: str = Query(..., min_length=1),

                   fmt: Format = Query(Format.CSV),
                   cache_get: bool = Query(True), cache_expire: int = Query(86400, lt=86400 * 15),
                   async_remote: bool = Query(True),
                   user: str = Depends(get_current_user),
                   ):
    """复杂参数函数。通过文件上传二进制方式。不可跨语言。"""
    return await _do(**locals(), **deserialize(file))


async def _do(request: Request,
              func: str,

              args: List[Any] = [],
              kwargs: Dict[str, Any] = {},

              fmt: Format = Format.CSV,
              cache_get: bool = True, cache_expire: int = 86400,
              async_remote: bool = True,
              file: bytes = None,  # 用于兼容文件上传模式，但实际没有使用
              user: str = None,  # 没有用到，用于token认证
              ):
    """实际处理函数"""
    try:
        before_call(request.client.host, user, func)
        key, buf, data = await call(user, func, args, kwargs, cache_get, cache_expire, async_remote)
    except Exception as e:
        # 主要是处理
        key = type(e).__name__
        # 这里没有缓存，因为这个错误是服务器内部检查
        data = RspModel(status=status.HTTP_401_UNAUTHORIZED,
                        datetime=datetime.now().isoformat(),
                        func=func, args=args, kwargs=kwargs)
        data.type = type(e).__name__
        data.data = repr(e)
        data = data.dict()
        buf = serialize(data).read()

    # 直接二进制返回
    if fmt == Format.PKL_GZIP:
        r = StreamingResponse(iter([buf]), media_type="application/octet-stream")
        content_disposition = f'attachment; filename="{key}.{Format.PKL_GZIP}"'
        r.headers.setdefault("content-disposition", content_disposition)
        return r

    # 从缓存中获取的，但又要转成json等其它功能，需要解码
    if data is None:
        data = deserialize(buf)

    # DataFrame需要csv格式时走此路径
    if fmt == Format.CSV and data['type'] in ('DataFrame', 'Series'):
        # 指定返回的Content-Type，这样MATLAB的webread可自动识别
        r = PlainTextResponse(data['data'].to_csv(), media_type="text/csv")
        return r

    # JSON格式时，为了方便DataFrame的显示，做一下转换
    data['data'] = obj_to_dict(data['data'])

    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-03-02

"""
API输入时间参数过于动态不利于缓存key生成

例如date.now()时间一直在变化，如果转成日期或转成周五，可以大量减少key
"""

from datetime import datetime

import numpy as np
import pandas as pd


def to_date_str(dt):
    """只取日期部分"""
    if isinstance(dt, int):
        dt = str(dt)
    if isinstance(dt, str):
        if ':' in dt:
            return dt[:10]
        if len(dt) == 10:
            return dt
    if isinstance(dt, (str, np.datetime64)):
        dt = pd.to_datetime(dt)
    return f'{dt:%Y-%m-%d}'


def eq_today(end_date):
    """等于今天"""
    return to_date_str(end_date) == f'{datetime.today():%Y-%m-%d}'


def ge_today(end_date):
    """大于等于今天"""
    return to_date_str(end_date) >= f'{datetime.today():%Y-%m-%d}'


if __name__ == '__main__':
    print(eq_today(20210925))
    print(eq_today(20210929))
    print(to_date_str(20210925))
    print(to_date_str('2021-09-25 01:44:29.477371'))
    print(to_date_str('2021-09-25T01:44:29.477371'))
    print(eq_today('2021-09-29T01:44:29.477371'))
    print(eq_today('2021-09-30'))
    print(ge_today('2021-09-30'))
    print(to_date_str(datetime.now()))
    print(to_date_str(datetime.now().date()))
    print(to_date_str(pd.to_datetime(datetime.now())))
    print(to_date_str(np.datetime64(datetime.now())))

    from dateutil.relativedelta import relativedelta, FR

    print(datetime(2021, 9, 25) + relativedelta(weekday=FR(-1)))
    print(datetime(2021, 9, 25) + relativedelta(weekday=FR(0)))
    print(datetime(2021, 9, 25) + relativedelta(weekday=FR(+1)))



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-05-26
"""
reverse rpc，反弹rpc
如果要调用的服务器在内网，并且由于条件所限，无法安装更多的库
所以这里只需小规模的安装即可

以下代码可以直接复制在Notebook中使用
"""

import asyncio

from loguru import logger

from ksrpc.connections.websocket import WebSocketConnection

TOKEN = 'secret-token-2'
URL = 'ws://127.0.0.1:8000/ws/client?room=9527'


async def async_main():
    while True:
        logger.info('try to connect')
        try:
            # 连接被控端地址（带/client），同时还要指定约定的房间号
            async with WebSocketConnection(URL, token=TOKEN) as conn:
                logger.info('connected')
                await conn.reverse()
        except Exception as e:
            logger.info(e)
            # 重试不能太块，一直在重试时，试着点重启内核
            await asyncio.sleep(5)


if __name__ == '__main__':
    asyncio.run(async_main())



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-05-22
"""
HTTP/WebSocket服务端入口
"""
import uvicorn

# HTTP服务
from ksrpc.app.http_ import *
# WebSocket服务
from ksrpc.app.websocket_ import *

if __name__ == "__main__":
    uvicorn.run(app, host='0.0.0.0', port=8000)
    # 添加几个上面库中的函数，防止在IDE在Reformat时自动删除import
    api_file
    websocket_endpoint_bytes



import setuptools

with open("README.md", "r", encoding='utf-8') as fp:
    long_description = fp.read()

version = {}
with open("ksrpc/_version.py", encoding="utf-8") as fp:
    exec(fp.read(), version)

setuptools.setup(
    name="ksrpc",
    version=version['__version__'],
    author="wukan",
    author_email="wu-kan@163.com",
    description="Keep Simple RPC",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wukan1986/ksrpc",
    packages=setuptools.find_packages(),
    # 默认。用于支持反弹RPC
    install_requires=[
        'loguru',
        'nest-asyncio',
        'pandas',
        'pydantic',
        'websockets',
    ],
    extras_require={
        # 服务端
        'server': [
            'aioredis',
            'fakeredis',
            'fastapi',
            'python-multipart',
            'uvicorn[standard]',
            'IPy',
        ],
        # 客户端
        'client': [
            'httpx',
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
    ],
    python_requires=">=3.6",
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-05-09
"""
调用器
根据指定参数，调用，并返回结果
"""

import hashlib
from datetime import datetime

import numpy as np
import pandas as pd
# from fastapi import status
from IPy import IP
from loguru import logger

from .cache import async_cache_get, async_cache_setex, async_cache_incrby
from .config import QUOTA_CHECK
from .model import RspModel
from .serializer.pkl_gzip import serialize
from .utils.async_ import to_async, to_sync
from .utils.check_ import check_methods, check_ip, get_quota


def make_key(func, args, kwargs):
    """生成缓存key"""
    args_kwargs = f'{repr(args)}_{repr(kwargs)}'
    if len(args_kwargs) > 32:
        args_kwargs = hashlib.md5(args_kwargs.encode('utf-8')).hexdigest()
    return f'{func}_{args_kwargs}'


def before_call(host, user, func):
    """调用前的检查"""
    from .config import (
        METHODS_CHECK, METHODS_ALLOW, METHODS_BLOCK,
        IP_CHECK, IP_ALLOW, IP_BLOCK
    )

    if IP_CHECK:
        host = IP(host)
        if not check_ip(IP_ALLOW, host, False):
            raise Exception(f'IP Not Allowed, {host} not in allowlist')
        if not check_ip(IP_BLOCK, host, True):
            raise Exception(f'IP Not Allowed, {host} in blocklist')

    if user is None:
        raise Exception(f"Unauthorized")

    methods = func.split('.')

    if METHODS_CHECK:
        if not check_methods(METHODS_ALLOW, methods, False):
            raise Exception(f'Method Not Allowed, {func} not in allowlist')
        if not check_methods(METHODS_BLOCK, methods, True):
            raise Exception(f'Method Not Allowed, {func} in blocklist')


async def call(user, func, args, kwargs, cache_get, cache_expire, async_remote, fmt=None):
    try:
        buf = None
        data = None  # None表示从缓存中取的，需要其它格式时需要转换
        # cache所用的key
        key = make_key(func, args, kwargs)

        if cache_get:
            # 优先取缓存
            buf = await async_cache_get(key)
        if buf is None:
            cache_expire, buf, data = await _call(user, func, args, kwargs, cache_expire, async_remote)

            # 过短的缓存时间没有必要
            if cache_expire >= 30:
                # 就算强行去查询数据，也会想办法存下，再次取时还能从缓存中取
                await async_cache_setex(key, cache_expire, buf)
                logger.info(f'To cache: expire:{cache_expire}\tlen:{len(buf)}\t{key}')
        else:
            logger.info(f'From cache: {func}\t{args}\t{kwargs}')

    except Exception as e:
        key = type(e).__name__
        # 这里没有缓存，因为这个错误是服务器内部检查
        data = RspModel(status=403,  # status.HTTP_403_FORBIDDEN,
                        datetime=datetime.now().isoformat(),
                        func=func, args=args, kwargs=kwargs)
        data.type = type(e).__name__
        data.data = repr(e)
        data = data.dict()
        buf = serialize(data).read()

    return key, buf, data


async def _call(user, func_name, args, kwargs, cache_expire, async_remote):
    """进行实际调用

    将配额计算放在调用第三方代码时，优化用户体验
    """

    methods = func_name.split('.')
    module = methods[0]

    m_quota_key = f'QUOTA/{user}/{module}'
    f_quota_key = f'QUOTA/{user}/{func_name}'
    if QUOTA_CHECK:
        # 配额检查
        from .config import QUOTA_MODULE, QUOTA_MODULE_DEFAULT, QUOTA_FUNC, QUOTA_FUNC_DEFAULT

        m_quota = int(await async_cache_get(m_quota_key) or 0)
        f_quota = int(await async_cache_get(f_quota_key) or 0)
        quota = get_quota(QUOTA_FUNC, methods, QUOTA_FUNC_DEFAULT)
        if f_quota > quota:
            raise Exception(f'Over quota: user:{user}, {func_name}:{f_quota}>{quota}')
        quota = get_quota(QUOTA_MODULE, [module], QUOTA_MODULE_DEFAULT)
        if m_quota > quota:
            raise Exception(f'Over quota: user:{user}, {module}:{quota}')

    try:
        # 当前server目录下文件，用于特别处理
        api = __import__(f'{__package__}.server.{module}', fromlist=['*'])
    except ModuleNotFoundError as e:
        # 导入系统包
        api = __import__(module, fromlist=['*'])

    # 返回的数据包
    d = RspModel(status=200,  # status.HTTP_200_OK,
                 datetime=datetime.now().isoformat(),  # 加查询时间，缓存中也许可以判断是否过期
                 func=func_name,
                 args=args,
                 kwargs=kwargs)

    methods = methods[1:]
    try:
        # 转成字符串，后面可能于做cache的key
        logger.info(f'Call: {func_name}\t{args}\t{kwargs}')

        func = api
        for method in methods:
            func = getattr(func, method)

        # 可以调用的属性
        if callable(func):
            # 例如os.remove
            type_ = 'function'
            if async_remote:
                func = await to_async(func)(*args, **kwargs)
            else:
                func = to_sync(func)(*args, **kwargs)

        # 结果的类型名
        class_name = func.__class__.__name__

        # 不可调用的直接返回
        if class_name == 'module':
            # 例如os.path，但用法为os.path()
            data = repr(func)
            cache_expire = min(cache_expire, 0)
        elif class_name in ('dict_keys', 'dict_values'):
            # 无法序列化，只能强行转换
            # 例如：sys.modules.keys(), 也可以sys.modules.keys.__list()
            data = list(func)
            cache_expire = min(cache_expire, 0)
        else:
            # 例如math.pi，但用法math.pi()
            data = func

            if QUOTA_CHECK:
                # 模块和函数都进行配额统计，是按行记（聚宽），按单元格记（万得），还是按调用次数记？
                if isinstance(data, (pd.DataFrame, pd.Series, np.ndarray, list)):
                    # # 按调用次数
                    # await async_cache_incrby(m_quota_key, 1)
                    # await async_cache_incrby(f_quota_key, 1)

                    # 按行统计
                    await async_cache_incrby(m_quota_key, len(data))
                    await async_cache_incrby(f_quota_key, len(data))

                    # # 按单元格统计
                    # if hasattr(data, 'size'):
                    #     await async_cache_incrby(m_quota_key, data.size)
                    #     await async_cache_incrby(f_quota_key, data.size)
                    # else:
                    #     await async_cache_incrby(m_quota_key, len(data))
                    #     await async_cache_incrby(f_quota_key, len(data))

        d.type = type(func).__name__
        d.data = data

        # pkl序列化，为了能完整还源
        d = d.dict()
        buf = serialize(d).read()

    except Exception as e:
        d.status = 500,  # status.HTTP_500_INTERNAL_SERVER_ERROR
        d.type = type(e).__name__
        d.data = repr(e)
        # 错误也缓存一会，防止用户写了死循环搞崩上游
        cache_expire = min(cache_expire, 60)
        d = d.dict()
        # 由于错误信息也想缓存，所以这里进行编码
        buf = serialize(d).read()

    return cache_expire, buf, d



"""
服务端配置
"""
from IPy import IP

"""
是否检查可调用的方法。
!!! 生产环境中请不要随意关闭，否则权限过大
"""
METHODS_CHECK = False

# 允许的方法
# 没有列出的module和method默认值为False表示不可调用
METHODS_ALLOW = {
    'pandas': {
        'read_csv': True,
        'util': {
            'testing': {
                'makeTimeDataFrame': True,
                'makeDataFrame': False,
            }
        }
    },
    'os': True,
    'ntpath': True,
    'numpy': True,
    'math': True,
    'sys': True,
    'jqdatasdk': True,
    'WindPy': True,
    'tushare': True,
    'rqdatac': True,
    'demo': True,
}

# 禁止的方法
# 没有列出的module和method默认值为True表示放行
METHODS_BLOCK = {
    'os': {
        'remove': False,
    },
    'jqdatasdk': {
        'is_auth': False,
        'auth': False,
        'logout': False,
    },
    'WindPy': {
        'w': {
            'start': False,
        }
    },
    'tushare': {
        'set_token': False,
        'pro_api': False,
    },
    'rqdatac': {
        'init': False,
    },
    'demo': True,
}

"""
是否进行IP检查，可屏蔽外网访问
"""
IP_CHECK = True

IP_ALLOW = {
    IP('127.0.0.0/8'): True,  # 回环地址
    IP('192.168.0.0/16'): True,  # C类内网
    IP('172.16.0.0/12'): True,  # B类内网
    IP('10.0.0.0/8'): True,  # A类内网
}
IP_BLOCK = {
    IP('8.8.8.8/32'): False,
    IP('8.8.4.4/32'): False,
}

"""
是否启用授权
"""
AUTH_CHECK = False
# API授权
AUTH_TOKENS = {
    "secret-token-1": "john",
    "secret-token-2": "susan",
}

"""
是否进行配额检查，可限制用户超量下载
只限制初始请求数据，第二次请求走缓存消耗配额
"""
QUOTA_CHECK = False

QUOTA_MODULE = {
    'demo': 1000,
}

QUOTA_FUNC = {
    'demo': {
        'test': 500,
    }
}
QUOTA_MODULE_DEFAULT = 10000 * 10000  # 1亿行
QUOTA_FUNC_DEFAULT = 10000 * 1000  # 1千万行

# 缓存类型。生产环境请配置redis服务
CACHE_TYPE = 'fakeredis'  # fakeredis, aioredis
# 缓存服务地址
REDIS_URL = "redis://:password@localhost:6379/0"

# 聚宽账号
JQ_USERNAME = '13912345678'
JQ_PASSWORD = '12345678'

# TuShare账号
TUSHARE_TOKEN = '12345678'

# eikon账号
EIKON_APP_KEY = '12345678'

# 米筐账号
RQ_USERNAME = '13912345678'
RQ_PASSWORD = '12345678'

# 启用万得
WIND_ENABLE = False
